package main

import (
	"encoding/json"
	"fmt"
	log "git.randomchars.net/FreeNitori/Log"
	"github.com/bwmarrin/discordgo"
	"os"
	"sync"
	"time"
)

var ticketLock = sync.RWMutex{}
var ticketInstancesChannel = make(map[string]*ticket)
var ticketInstancesUser = make(map[string]*ticket)

type ticket struct {
	ChannelID     string    `json:"channel_id"`
	UserID        string    `json:"user_id"`
	UserChannelID string    `json:"user_channel_id"`
	CreationDate  time.Time `json:"creation_date"`
}

type ticketState struct {
	TicketInstancesChannel map[string]*ticket `json:"ticket_instances_channel"`
	TicketInstancesUser    map[string]*ticket `json:"ticket_instances_user"`
}

func (t *ticket) delete() {
	ticketLock.RLock()
	defer ticketLock.RUnlock()

	delete(ticketInstancesUser, t.UserID)
	delete(ticketInstancesChannel, t.ChannelID)
	if _, err := session.ChannelDelete(t.ChannelID); err != nil {
		log.Warnf("Error while deleting channel %s, %s", t.ChannelID, err)
	}
	sendMessage(t.UserChannelID, "Conversation has ended.")
}

func createTicket(channelID, userID string) *ticket {
	_, ok := ticketInstancesUser[userID]
	if ok {
		return nil
	}

	instance := &ticket{
		ChannelID:    channelID,
		UserID:       userID,
		CreationDate: time.Now(),
	}

	if userChannel, err := session.UserChannelCreate(userID); err != nil {
		log.Warnf("Error creating user channel for user %s, %s", userID, err)
		return nil
	} else {
		instance.UserChannelID = userChannel.ID
	}
	ticketInstancesChannel[channelID] = instance
	ticketInstancesUser[userID] = instance
	log.Infof("Ticket created for user %s in channel %s.", userID, channelID)
	return instance
}

func setupTicket(guildID, userID string, subject int, user *discordgo.User) bool {
	ticketLock.RLock()
	defer ticketLock.RUnlock()
	if channel, err := session.GuildChannelCreate(guildID, "ticket-"+userID, discordgo.ChannelTypeGuildText); err != nil {
		log.Warnf("Error creating ticket channel for user %s, %s", userID, err)
		return false
	} else {
		if instance := createTicket(channel.ID, userID); instance == nil {
			return false
		} else {
			if !sendMessage(instance.UserChannelID,
				fmt.Sprintf("Ticket opened with subject %v. Start your conversation here.",
					subject)) {
				return false
			}
		}
		if user != nil {
			return sendMessage(channel.ID, fmt.Sprintf("Ticket created by user %s#%s (%s) with subject %v.",
				user.Username, user.Discriminator, user.ID, subject))
		} else {
			return sendMessage(channel.ID, fmt.Sprintf("Ticket created with subject %v.", subject))
		}
	}
}

func makeActiveTicketsPayload() []byte {
	if payload, err := json.Marshal(ticketState{
		TicketInstancesChannel: ticketInstancesChannel,
		TicketInstancesUser:    ticketInstancesUser,
	}); err != nil {
		log.Fatalf("Error saving ticket state, %s, all ticket changes since previous start will be lost.")
		os.Exit(1)
	} else {
		return payload
	}
	return nil
}

func dumpTicketState(path string) {
	// Permanently lock ticket state until program exit
	ticketLock.Lock()

	if err := os.WriteFile(path, makeActiveTicketsPayload(), 0600); err != nil {
		log.Errorf("Error dumping ticket state, %s, all ticket changes since previous start will be lost.")
	}
}

func readTicketState(path string) (ticketState, bool) {
	var state ticketState
	if payload, err := os.ReadFile(path); err != nil {
		if os.IsNotExist(err) {
			return state, false
		}
		log.Fatalf("Error reading state file, %s", err)
		os.Exit(1)
	} else {
		if err = json.Unmarshal(payload, &state); err != nil {
			log.Fatalf("Error parsing state file, %s", err)
			os.Exit(1)
		} else {
			return state, true
		}
	}
	return state, false
}
